{$I UOpt.INC}				(* Standard Conditional-Defines *)
UNIT Canlines;

INTERFACE

USES SysUtils, WinProcs, WinTypes, Classes;

TYPE
  TWhatAction   = ( waNone, waCreate, waDelete, waModify, waRestore );
  TSource	= ( srcNone, srcPresent );
  TFileState	= ( fstNone, fstLoaded, fstSaved, fstNew, fstModified );

  TIsOnline = RECORD
    bYes    : BOOLEAN;
    dEps    : DOUBLE;
    END;

  TMyClassesRec= RECORD
    lTimeStamp : LONGINT;
    Index      : INTEGER;
    aWA        : TWhatAction;
    bSelected  : BOOLEAN;
    END;

  TMyClasses   = CLASS( TPersistent )
   PUBLIC
    Def        : TMyClassesRec;
    END;

  TNodeRec      = PACKED RECORD
    thePoint	: TPoint;
    wZL		: WORD;
    strName	: ShortString;
    bShowName	: BOOLEAN;
    theSource	: TSource;
    END;

  TNode 	= CLASS( TMyClasses )
   PUBLIC
    Data        : TNodeRec;
    END;

  TTrmLineRec   = PACKED RECORD
    tpStart	: TPoint;
    tpStop	: TPoint;
    wZL		: WORD;
    wLength	: WORD;
    sAngle      : SINGLE;
    strName	: ShortString;
    bShowLength	: BOOLEAN;
    END;

  TTrmLine	= CLASS( TMyClasses )
   PUBLIC
    Data        : TTrmLineRec;
    END;

TYPE
  TCanNodes	= CLASS( TStringList )
   PRIVATE
    strActualFile	: TFileName;
    tMyFileState	: TFileState;
    tpMax		: TPoint;
    tpMin		: TPoint;
    wSources		: WORD;
    wNodes		: WORD;
    wLines		: WORD;
    wDefaultNoZL	: WORD;
    wDefaultZL		: WORD;
    wDefaultLineZL	: WORD;
    slUndo              : TStringList;
    lTickCountUpdate    : LONGINT;
    FUNCTION RemoveNodeAt2( thePoint: TPoint ): BOOLEAN;
    FUNCTION BuildLineCoordinateString( aStartPoint, aStopPoint: TPoint ): ShortString;
    FUNCTION BuildNodeCoordinateString( aPoint: TPoint ): ShortString;
    FUNCTION AddNode( aNode: TNode ): BOOLEAN;
    FUNCTION AddTrmLine( aTrmLine: TTrmLine ): BOOLEAN;
    FUNCTION SplitLineAt( aPoint: TPoint ): BOOLEAN;
    FUNCTION PointIsOnLine( aPoint: TPoint; aTrmLine: TTrmLine ): TIsOnline;

  FUNCTION ChangeNodeZLAt( aPoint : TPoint; wZL: WORD ): BOOLEAN;
  FUNCTION ChangeNodeSourceAt( aPoint : TPoint; aSource: TSource ): BOOLEAN;

    FUNCTION ClearStringList( aStringList: TStringlist ): BOOLEAN;
    FUNCTION DeleteObject( iIndex: INTEGER ): BOOLEAN;
   PROTECTED
   PUBLIC
    CONSTRUCTOR Create;
    CONSTRUCTOR CreateCANLines( wDefZL, wDefNoZL, wDefLineZL: WORD );
    DESTRUCTOR  Destroy;		                   		OVERRIDE;
    FUNCTION EnumerateNodes: INTEGER;
    FUNCTION ShiftNetwork( theIncrement: TPoint ): BOOLEAN;

    FUNCTION AddObject( CONST S: STRING; AObject: TObject ): INTEGER;	OVERRIDE;

    FUNCTION Load( strFileName: TFileName ): BOOLEAN;
    FUNCTION Save( strFileName: TFileName ): BOOLEAN;
    FUNCTION SaveCircuit( strFileName: TFileName ): BOOLEAN;

    FUNCTION AddNodeAt( aPoint: TPoint ): BOOLEAN;
    FUNCTION NodeExistAt( aPoint: TPoint ): BOOLEAN;
    FUNCTION GetNodeAt( aPoint: TPoint ): TNode;
    FUNCTION GetNodeNameAt( aPoint: TPoint ): ShortString;
    FUNCTION ModifyingTrmLine( aTrmLine: TTrmLine ): BOOLEAN;
    FUNCTION ModifyingNode( aNode: TNode ): BOOLEAN;
    FUNCTION RemoveNodeAt( aPoint: TPoint ): BOOLEAN;

    FUNCTION AddTrmLineAt( firstPoint, lastPoint: TPoint; wDefZL: WORD; bShL: BOOLEAN ): BOOLEAN;
    FUNCTION TrmLineExists( aStartPoint, aStopPoint: TPoint ): BOOLEAN;
    FUNCTION GetTrmLineAt( aPoint: TPoint ): TTrmLine;
    FUNCTION GetTrmLineAt2( aStartPoint, aStopPoint: TPoint ): TTrmLine;
    FUNCTION RemoveTrmLine( VAR theTrmLine: TTrmline ): BOOLEAN;

    FUNCTION AddCircuitToStringList( aStringList: TStringlist ): BOOLEAN;
    FUNCTION UnDo: BOOLEAN;
    FUNCTION UnDoPossible: BOOLEAN;
    FUNCTION NetworkExists: BOOLEAN;
   PUBLISHED
    PROPERTY FFileName	: TFileName	READ strActualFile;
    PROPERTY FFileState : TFileState	READ tMyFileState;
    PROPERTY FNumSources : WORD		READ wSources;
    PROPERTY FNumNodes: WORD		READ wNodes;
    PROPERTY FNumLines: WORD		READ wLines;
    PROPERTY FNoZL : WORD		READ wDefaultNoZL	WRITE wDefaultNoZL;
    PROPERTY FZL: WORD			READ wDefaultZL		WRITE wDefaultZL;
    PROPERTY FLineZL: WORD	  	READ wDefaultLineZL 	WRITE wDefaultLineZL;
    PROPERTY FTickCount: LONGINT        READ lTickCountUpdate   WRITE lTickCountUpdate;
    PROPERTY FMax: TPoint	        READ tpMax;
    PROPERTY FMin: TPoint       	READ tpMin;
    PROPERTY FUnDoPossible: BOOLEAN     READ UnDoPossible;
    PROPERTY FNetworkExists: BOOLEAN    READ NetworkExists;
    END;


IMPLEMENTATION

(*--------------------------------------------------------*)
FUNCTION LMin( l1, l2: LONGINT): LONGINT;

BEGIN
IF l1 < l2 THEN LMin := l1 ELSE LMin := l2;
END;

(*--------------------------------------------------------*)
FUNCTION LMax( l1, l2: LONGINT): LONGINT;

BEGIN
IF l1 < l2 THEN LMax := l2 ELSE LMax := l1;
END;

(*--------------------------------------------------------*)
FUNCTION LSign( l1: LONGINT ): EXTENDED;

BEGIN
IF l1 < 0 THEN Result := -1.0 ELSE Result := 1.0;
END;

(*--------------------------------------------------------*)
FUNCTION PointInRect( aP1Rect, aP2Rect, aPoint: TPoint ): BOOLEAN;

BEGIN
IF (aPoint.X >= LMin(aP1Rect.X, aP2Rect.X)) AND (aPoint.X <= LMax(aP1Rect.X, aP2Rect.X)) AND
   (aPoint.Y >= LMin(aP1Rect.Y, aP2Rect.Y)) AND (aPoint.Y <= LMax(aP1Rect.Y, aP2Rect.Y)) THEN
   PointInRect := true
  ELSE
   PointInRect := false;
END;

(*--------------------------------------------------------*)
FUNCTION F2Str( fIn: EXTENDED; wDec: WORD ): ShortString;

VAR
  strF	: STRING[40];
  wEPos : WORD;

BEGIN
IF wDec > 0 THEN
  BEGIN
  Str( fIn:(9+wDec), strF );
  WHILE strF[1] = ' ' DO
   Delete( strF, 1, 1 );
  wEPos := Pos( 'E', strF );
  IF wEPos > 0 THEN
    BEGIN
    IF strF[wEPos+1] = '-' THEN
      BEGIN
      WHILE (strF[wEPos+2] = '0') AND (Length(strF) > wEPos+2) DO
	Delete( strF, wEPos+2, 1 );
      END;
    IF strF[wEPos+1] = '+' THEN
      BEGIN
      Delete( strF, wEPos+1, 1 );
      WHILE (strF[wEPos+1] = '0') AND (Length(strF) > wEPos+1) DO
	Delete( strF, wEPos+1, 1 );
      END;
    END;
  END
 ELSE
  BEGIN
  Str( fIn:5:0, strF );
  END;
F2Str := strF;
END;

(*--------------------------------------------------*)
FUNCTION IsSamePoint( aPoint0, aPoint1: TPoint ): BOOLEAN;

BEGIN
Result := (aPoint0.X = aPoint1.X) AND (aPoint0.Y = aPoint1.Y);
END;


(*--------------------------------------------------*)
CONSTRUCTOR TCanNodes.Create;

BEGIN
INHERITED Create;
RegisterClass( TNode );
RegisterClass( TTrmLine );
strActualFile	:= 'newfile.rpc';
tMyFileState	:= fstNew;
Duplicates 	:= dupError;
Sorted		:= true;
wSources	:= 0;
wNodes		:= 0;
wLines		:= 0;
wDefaultNoZL	:= 65535;
wDefaultZL	:= 120;
wDefaultLineZL  := 120;
tpMax.X     	:= -2147483647;
tpMax.Y     	:= -2147483647;
tpMin.X		:=  2147483647;
tpMin.Y     	:=  2147483647;
slUndo          := TStringlist.Create;
WITH slUndo DO
  BEGIN
  Sorted        := false;
  Duplicates    := dupAccept;
  END;

END;

(*--------------------------------------------------*)
CONSTRUCTOR TCanNodes.CreateCANLines( wDefZL, wDefNoZL, wDefLineZL: WORD );

BEGIN
Create;
wDefaultNoZL	:= wDefNoZL;
wDefaultZL	:= wDefZL;
wDefaultLineZL  := wDefLineZL;
END;

(*--------------------------------------------------*)
DESTRUCTOR  TCanNodes.Destroy;

BEGIN
ClearStringList( Self  );
ClearStringList( slUndo );
INHERITED Destroy;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.BuildNodeCoordinateString( aPoint: TPoint ): ShortString;

BEGIN
BuildNodeCoordinateString :=
	'N'+IntToHex(aPoint.X, 4)+':'+IntToHex(aPoint.Y, 4);
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.BuildLineCoordinateString( aStartPoint, aStopPoint: TPoint ): ShortString;

BEGIN
BuildLineCoordinateString :=
	'W'+IntToHex(aStartPoint.X, 4)+
	':'+IntToHex(aStartPoint.Y, 4)+
	'::'+IntToHex(aStopPoint.X, 4)+
	':'+IntToHex(aStopPoint.Y, 4);

END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.EnumerateNodes: INTEGER;

VAR
  iCount 	: INTEGER;
  aNode		: TNode;

BEGIN
IF Count > 0 THEN
  BEGIN
  FOR iCount := 0 TO Count-1 DO
    BEGIN
    IF Objects[ iCount ] IS TNode THEN
      BEGIN
      aNode	:= TNode(Objects[ iCount ]);
      IF Pos(']', aNode.Data.strName) <> 0 THEN
	System.Delete( aNode.Data.strName, 1, Pos(']', aNode.Data.strName)+1 );
      aNode.Data.strName := '['+IntToStr( IndexOfObject( aNode )+1 )+'] '+aNode.Data.strName;
      END;
    END;
  END;
Result := Count;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.ShiftNetwork( theIncrement: TPoint ): BOOLEAN;

VAR
  iCount 	: INTEGER;
  aNode		: TNode;
  aTrmLine      : TTrmLine;

BEGIN
IF Count > 0 THEN
  BEGIN
  tpMax.X     	:= -2147483647;
  tpMax.Y     	:= -2147483647;
  tpMin.X	:=  2147483647;
  tpMin.Y     	:=  2147483647;
  BeginUpdate;
  Duplicates 	:= dupAccept;
  Sorted	:= false;
  FOR iCount := 0 TO Count-1 DO
    BEGIN
    IF Objects[ iCount ] IS TNode THEN
      BEGIN
      aNode	:= TNode(Objects[ iCount ]);
      WITH aNode.Data DO
	BEGIN
	Inc( thePoint.X, theIncrement.X );
	Inc( thePoint.Y, theIncrement.Y );
	Strings[ iCount ] := BuildNodeCoordinateString( thePoint );
	END; //WITH aNode.Data
      WITH tpMax DO
	BEGIN
	X     := LMax(X, aNode.Data.thePoint.X );
	Y     := LMax(Y, aNode.Data.thePoint.Y );
	END;
      WITH tpMin DO
	BEGIN
	X     := LMin(X, aNode.Data.thePoint.X );
	Y     := LMin(Y, aNode.Data.thePoint.Y );
	END;
      END; //IF Objects[ iCount ]
    IF Objects[ iCount ] IS TTrmLine THEN
      BEGIN
      aTrmLine 	:= TTrmLine(Objects[ iCount ]);
      WITH aTrmLine.Data DO
	BEGIN
	Inc( tpStart.X, theIncrement.X );
	Inc( tpStop.X, theIncrement.X );
	Inc( tpStart.Y, theIncrement.Y );
	Inc( tpStop.Y, theIncrement.Y );
	Strings[ iCount ] := BuildLineCoordinateString( tpStart, tpStop );
	END; //WITH aNode.Data
      END; //IF Objects[ iCount ]
    END; //FOR iCount := 0
  Duplicates 	:= dupError;
  Sorted	:= true;
  EndUpdate;
  END; //IF Count
Result := Count > 0;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.AddObject( CONST S: STRING; AObject: TObject ): INTEGER;

VAR
  aNode         : TNode;
  aTrmLine      : TTrmLine;

BEGIN
Result := INHERITED AddObject( S, AObject );
IF (Result >= 0) THEN
  BEGIN
  tMyFileState	:= fstModified;
  IF (AObject IS TNode) THEN
    BEGIN
    TNode(AObject).Def.aWA                := waCreate;
    aNode                                 := TNode.Create;
    aNode.Data                            := TNode(AObject).Data;
    aNode.Def.Index                       := Result;
    aNode.Def.lTimeStamp                  := FTickCount;
    aNode.Def.aWa                         := waCreate;
    Inc( wNodes );
    IF TNode(AObject).Data.theSource = srcPresent THEN
      Inc(wSources);
    slUndo.AddObject( S, TObject(aNode) );
    WITH tpMax DO
      BEGIN
      X     := LMax(X, aNode.Data.thePoint.X );
      Y     := LMax(Y, aNode.Data.thePoint.Y );
      END;
    WITH tpMin DO
      BEGIN
      X     := LMin(X, aNode.Data.thePoint.X );
      Y     := LMin(Y, aNode.Data.thePoint.Y );
      END;
    END
   ELSE
    IF (AObject IS TTrmLine) THEN
      BEGIN
      Inc( wLines );
      TTrmLine(AObject).Def.aWA           := waCreate;
      aTrmLine                            := TTrmLine.Create;
      aTrmLine.Data                       := TTrmLine(AObject).Data;
      aTrmLine.Def.Index                  := Result;
      aTrmLine.Def.lTimeStamp             := FTickCount;
      aTrmLine.Def.aWA                    := waCreate;
      slUndo.AddObject( S, TObject(aTrmLine) );
      END;
  END;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.AddCircuitToStringList( aStringList: TStringlist ): BOOLEAN;

VAR
  iCount 	: INTEGER;
  aNodeStart	: TNode;
  aNodeStop	: TNode;
  aNode		: TNode;
  aTrmLine	: TTrmLine;
  strLine	: ShortString;
  iNStart	: INTEGER;
  iNStop	: INTEGER;

BEGIN
IF Count > 0 THEN
  BEGIN
  FOR iCount := 0 TO Count-1 DO				{zuerste Leitungen}
    BEGIN
    IF Objects[ iCount ] IS TTrmLine THEN
      BEGIN
      aTrmLine	:= TTrmLine( Objects[ iCount ] );
      aNodeStart:= GetNodeAt( aTrmLine.Data.tpStart );
      aNodeStop	:= GetNodeAt( aTrmLine.Data.tpStop );
      iNStart	:= IndexOfObject( aNodeStart );
      iNStop	:= IndexOfObject( aNodeStop );
      strLine	:= '';
      aStringList.Add( strLine );
      strLine	:= '* Line '+aNodeStart.Data.strName+' <=> '+aNodeStop.Data.strName;
      aStringList.Add( strLine );
      strLine	:= '  T'+IntToStr( iNStart+1 )+IntToStr( iNStop+1 );				{T12}
      strLine	:=  strLine+' '+IntToStr( iNStart+1 )+' 0 '+IntToStr( iNStop+1 )+' 0 ';		{1 0 2 0 }
      strLine	:=  strLine+'Z0='+IntToStr( aTrmLine.Data.wZL );
      strLine	:=  strLine+' TD={'+F2Str( Int(aTrmLine.Data.wLength)*1.0E-3, 3 )+'/{xC0red}}';
      aStringList.Add( strLine );
      END;
    END;
  FOR iCount := 0 TO Count-1 DO				{jetzt Knoten}
    BEGIN
    IF Objects[ iCount ] IS TNode THEN
      BEGIN
      strLine	:= '';
      aStringList.Add( strLine );
      aNode	:= TNode(Objects[ iCount ]);
      iNStart	:= IndexOfObject( aNode );
      strLine	:= '* Node '+aNode.Data.strName;
      IF Pos(']', aNode.Data.strName) <> 0 THEN
	System.Delete( aNode.Data.strName, 1, Pos(']', aNode.Data.strName)+1 );
      aNode.Data.strName := '['+IntToStr( iNStart+1 )+'] '+aNode.Data.strName;
      aStringList.Add( strLine );
      strLine	:= '  RZL'+IntToStr( iNStart+1 )+' '+IntToStr( iNStart+1 )+' 0 '+IntToStr( aNode.Data.wZL );
      aStringList.Add( strLine );
      IF (aNode.Data.theSource = srcPresent) THEN
	BEGIN
	strLine	:= '  X'+IntToStr( iNStart+1 )+' '+IntToStr( iNStart+1 )+' CANDRV';
	aStringList.Add( strLine );
	END;
      END;
    END;
  strLine	:= '';
  aStringList.Add( strLine );
  END;
Result := Count > 0;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.SaveCircuit( strFileName: TFileName ): BOOLEAN;

VAR
  aStringList	: TStringList;

BEGIN
aStringList		:= TStringList.Create;
aStringList.Sorted	:= false;
aStringList.Duplicates	:= dupAccept;
Result			:= AddCircuitToStringList( aStringList );
aStringList.SaveToFile( strFileName );
aStringList.Free;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.Save( strFileName: TFileName ): BOOLEAN;

VAR
  iCount 	: INTEGER;
  aNode		: TNode;
  aTrmLine	: TTrmLine;
  aStream	: TStream;
  aWriter	: TWriter;

BEGIN
Result := false;
strActualFile := strFileName;
IF Count > 0 THEN
  BEGIN
  aStream := TFileStream.Create( strActualFile, fmCreate );
  aWriter := TWriter.Create( aStream, 2000 );
  aWriter.WriteListBegin;
  FOR iCount := 0 TO Count-1 DO
    BEGIN
    IF Objects[ iCount ] IS TNode THEN
      BEGIN
      aNode := TNode( Objects[ iCount ] );
      WITH aWriter, aNode DO
	BEGIN
	WriteString( ClassName );
	Write( Data, SizeOf(Data) );
	END;
      END
     ELSE
      IF Objects[ iCount ] IS TTrmLine THEN
	BEGIN
	aTrmLine := TTrmLine( Objects[ iCount ] );
	WITH aWriter, aTrmLine DO
	  BEGIN
	  WriteString( ClassName );
	  Write( Data, SizeOf(Data) );
	  END;
	END;
    END;
  aWriter.WriteListEnd;
  aWriter.Free;
  aStream.Free;
  tMyFileState	:= fstSaved;
  Result := true;
  END;

IF Result THEN
  ClearStringList( slUndo );
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.Load( strFileName: TFileName ): BOOLEAN;

VAR
  aStream	: TStream;
  aReader	: TReader;
  strClassName	: ShortString;
  ClassRef	: TClass;
  aObject	: TObject;


BEGIN
aStream	        := nil;
aReader	        := nil;
Result          := false;
wSources	:= 0;
wNodes		:= 0;
wLines		:= 0;
tpMax.X     	:= -2147483647;
tpMax.Y     	:= -2147483647;
tpMin.X		:=  2147483647;
tpMin.Y     	:=  2147483647;
IF NOT(FileExists( strFileName )) THEN
  BEGIN
  Save( strFileName );
  Exit;                                                 //!!! Ausgang !!!
  END;
TRY
  BeginUpdate;
  strActualFile := strFileName;
  aStream := TFileStream.Create( strActualFile, fmOpenRead );
  aReader := TReader.Create( aStream, 2000 );
  aReader.ReadListBegin;
  WHILE NOT aReader.EndOfList DO
    BEGIN
    strClassName	:= aReader.ReadString;
    ClassRef 		:= FindClass( strClassName );
    aObject		:= ClassRef.Create;
    IF aObject IS TNode THEN
      BEGIN
      WITH aReader, TNode(aObject) DO
	BEGIN
	Read( Data, SizeOf(Data) );
	END;
      AddNode( TNode(aObject) );
      END
     ELSE
      IF aObject IS TTrmLine THEN
	BEGIN
	WITH aReader, TTrmLine( aObject ) DO
	  BEGIN
	  Read( Data, SizeOf(Data) );
	  END;
	AddTrmLine( TTrmLine( aObject ) );
	END;
    Result := true;
    END; {WHILE NOT}
  aReader.ReadListEnd;
  aReader.Free;
  aStream.Free;
  tMyFileState	:= fstLoaded;
  EndUpdate;
  IF Result THEN
    ClearStringList( slUndo );
 EXCEPT
  IF Assigned(aReader) THEN
    aReader.Free;
  IF Assigned(aStream) THEN
    aStream.Free;
  Result := false;
  END;

END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.NodeExistAt( aPoint: TPoint ): BOOLEAN;

VAR
  iIndex	: INTEGER;

BEGIN
NodeExistAt := Find( BuildNodeCoordinateString( aPoint ), iIndex );
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.AddNode( aNode: TNode ): BOOLEAN;

BEGIN
Result := (AddObject( BuildNodeCoordinateString(aNode.Data.thePoint), aNode ) >= 0);
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.AddNodeAt( aPoint: TPoint ): BOOLEAN;

VAR
  aNewNode	: TNode;

BEGIN
IF NOT NodeExistAt( aPoint ) THEN
  BEGIN
  aNewNode		:= TNode.Create;
  WITH aNewNode.Data DO
    BEGIN
    thePoint            := aPoint;
    wZL	                := wDefaultNoZL;
    bShowName           := true;
    END;

  AddObject( BuildNodeCoordinateString( aPoint ), aNewNode );
  SplitLineAt( aPoint );
  END;
Result := (tMyFileState = fstModified);
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.TrmLineExists( aStartPoint, aStopPoint: TPoint ): BOOLEAN;

VAR
  iIndex	: INTEGER;

BEGIN
TrmLineExists := Find( BuildLineCoordinateString( aStartPoint, aStopPoint ), iIndex ) OR
		 Find( BuildLineCoordinateString( aStopPoint, aStartPoint ), iIndex );
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.AddTrmLine( aTrmLine: TTrmLine ): BOOLEAN;

VAR
  bLength	: BOOLEAN;

BEGIN
WITH aTRMLine.Data DO
  bLength := ((tpStop.X - tpStart.X) <> 0) OR ((tpStop.Y - tpStart.Y) <> 0);
IF bLength AND NOT( TrmLineExists( aTrmLine.Data.tpStart, aTrmLine.Data.tpStop ) ) THEN
  BEGIN
  Result := (AddObject( BuildLineCoordinateString(aTrmLine.Data.tpStart, aTrmLine.Data.tpStop ), aTrmLine ) >= 0);
  END
 ELSE
  BEGIN
  aTrmLine.Free;
  AddTrmLine := false;
  END;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.AddTrmLineAt( firstPoint, lastPoint: TPoint; wDefZL: WORD; bShL: BOOLEAN ): BOOLEAN;

VAR
  aNewTRMLine : TTRMLine;

BEGIN
aNewTRMLine := TTRMLine.Create;
WITH aNewTRMLine.Data DO
  BEGIN
  IF firstPoint.X < lastPoint.X THEN			{einsortieren nach "Grsse", zuerst X-Richtung}
    BEGIN
    tpStart	:= firstPoint;
    tpStop	:= lastPoint;
    END
   ELSE
    BEGIN
    IF firstPoint.X = lastPoint.X THEN			{senkrecht, jetzt nach Y}
      BEGIN
      IF firstPoint.Y < lastPoint.Y THEN		{hier gibts blo zwei Mglichkeiten, sonst Lnge = NULL}
	BEGIN
	tpStart	:= firstPoint;
	tpStop	:= lastPoint;
	END
       ELSE
	BEGIN
	tpStart	:= lastPoint;
	tpStop	:= firstPoint;
	END; {IF firstPoint.Y < lastPoint.Y}
      END
     ELSE
      BEGIN						{negative Steigung}
      tpStart	:= lastPoint;
      tpStop	:= firstPoint;
      END; {IF firstPoint.X = lastPoint.X}
    END; {IF firstPoint.X < lastPoint.X}

  wZL		:= wDefZL;
  bShowLength	:= bShL;
  TRY
    sAngle := ArcTan( Int(tpStop.Y-tpStart.Y) / Int(tpStop.X-tpStart.X) );
   EXCEPT
    ON EZeroDivide DO sAngle := 0.5*Pi; 			       	{senkrechte TrmLine}
    END;
  END;
Result := AddTrmLine( aNewTRMLine );
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.PointIsOnLine( aPoint: TPoint; aTrmLine: TTrmLine ): TIsOnline;

TYPE
  TLineType = (ltHorz, ltVert, ltOther);

VAR
  ldY           : LONGINT;
  ldX           : LONGINT;
  dAlpha1 	: DOUBLE;
  dAlpha2 	: DOUBLE;
  dBeta 	: DOUBLE;
  dB		: DOUBLE;
  aLineType     : TLineType;

BEGIN
WITH aTrmLine.Data DO
  BEGIN
  IF (tpStart.Y = tpStop.Y) THEN
    aLineType := ltVert
   ELSE
    IF (tpStart.X = tpStop.X) THEN
      aLineType := ltHorz
     ELSE
      aLineType := ltOther;
  CASE aLineType OF
    ltVert:
      BEGIN
      IF (tpStart.X < aPoint.X) AND (tpStop.X > aPoint.X) THEN
        BEGIN
        dB      := Abs( Int(aPoint.Y-tpStart.Y) );
        IF dB < 25.0 THEN
          Result.bYes	:= true
         ELSE
          Result.bYes	:= false;
        Result.dEps := dB;
        END
       ELSE
        BEGIN
        Result.bYes := false;
        Result.dEps := 1.0E400;
        END;
      END; {ltVert}
    ltHorz:
      BEGIN
      IF (tpStart.Y < aPoint.Y) AND (tpStop.Y > aPoint.Y) THEN
	BEGIN
	dB      := Abs( Int(aPoint.X-tpStart.X) );
        IF dB < 25.0 THEN
          Result.bYes	:= true
         ELSE
          Result.bYes	:= false;
        Result.dEps := dB;
        END
       ELSE
        BEGIN
        Result.bYes := false;
        Result.dEps := 1.0E400;
        END;
      END; {ltHorz}
    ltOther:
      BEGIN
      IF PointInRect( tpStart, tpStop, aPoint ) THEN
        BEGIN
        ldX := tpStop.X-tpStart.X;
        ldY := tpStop.Y-tpStart.Y;
        TRY
          dAlpha1 := ArcTan( Int(ldY) / Int(ldX) );
         EXCEPT
          ON EZeroDivide DO dAlpha1 := 0.5*Pi; 			       	{senkrechte TrmLine}
          END;

        ldX := aPoint.X-tpStart.X;
        ldY := aPoint.Y-tpStart.Y;
        TRY
          dBeta := ArcTan( Int(ldY) / Int(ldX) );
         EXCEPT
          ON EZeroDivide DO dBeta := 0.5*Pi*LSign(ldY);			{senkrechte Hilfslinie}
          END;

	dAlpha2 := dBeta-dAlpha1;
	dB      := Abs(Sqrt( Sqr(Int(ldY)) + Sqr(Int(ldX)) ) * Sin(dAlpha2));
        IF dB < 25.0 THEN
          Result.bYes	:= true
         ELSE
          Result.bYes	:= false;

        Result.dEps     := dB;
        END
       ELSE
        BEGIN
        Result.bYes     := false;
        Result.dEps     := 1.0E400;
	END; {IF PointInRect}
      END; {ltOther}
    END; {CASE}
  END; {WITH aTrmLine DO}
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.GetTrmLineAt2( aStartPoint, aStopPoint: TPoint ): TTrmLine;

VAR
  iIndex	: INTEGER;

BEGIN
IF Find( BuildLineCoordinateString( aStartPoint, aStopPoint ), iIndex ) THEN
  BEGIN
  Result := TTrmLine( Objects[ iIndex ] );
  END
 ELSE
  BEGIN
  Result := nil;
  END;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.GetTrmLineAt( aPoint: TPoint ): TTrmLine;

VAR
  iCount 	: INTEGER;
  aTrmLine	: TTrmLine;
  aTrmLineFound : TTrmLine;
  dMin          : DOUBLE;
  aIsOnLine	: TIsOnline;

BEGIN
dMin            := 1.0E400;
aTrmLineFound   := nil;
IF Count > 0 THEN
  BEGIN
  iCount 	:= 0;
  WHILE (iCount < Count) DO
    BEGIN
    IF Objects[ iCount ] IS TTrmLine THEN
      BEGIN
      aTrmLine		:= TTrmLine( Objects[ iCount ] );
      aIsOnLine		:= PointIsOnLine( aPoint, aTrmLine );
      WITH aIsOnLine DO
	BEGIN
	IF bYes THEN
	  BEGIN
	  IF dEps < dMin THEN
	    BEGIN
            dMin := dEps;
	    aTrmLineFound := aTrmLine;
	    END; {IF dEps < dMin}
	  END; {IF bYes}
	END; {WITH aIsOnLine}
      END; {IF Objects[ iCount ]}
    Inc( iCount );
    END; {WHILE (iCount < Count)}
  END; {IF Count > 0}
Result := aTrmLineFound;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.SplitLineAt( aPoint: TPoint ): BOOLEAN;

VAR
  iCount 	: INTEGER;
  aTrmLine	: TTrmLine;
  aPoint0	: TPoint;
  aPoint1	: TPoint;
  bPointOnLine	: BOOLEAN;
  bShL		: BOOLEAN;
  wDefZL	: WORD;


BEGIN				{NUR aufrufen, wenn nicht auf Knoten!}
bPointOnLine	:= false;
IF Count > 0 THEN
  BEGIN
  iCount 	:= 0;
  WHILE (iCount < Count) AND NOT(bPointOnLine) DO
    BEGIN
    IF Objects[ iCount ] IS TTrmLine THEN
      BEGIN
      aTrmLine		:= TTrmLine( Objects[ iCount ] );
      bPointOnLine	:= PointIsOnLine( aPoint, aTrmLine ).bYes;
      IF bPointOnLine THEN
	BEGIN
	WITH aTrmLine.Data DO                                   {tunlichts NICHT verndern!!!}
	  BEGIN
	  aPoint0	:= tpStart;				{alte Koordinaten merken}
	  aPoint1	:= tpStop;
	  bShL		:= bShowLength;
	  wDefZL	:= wZL;
	  DeleteObject( iCount );
	  AddTrmLineAt( aPoint0, aPoint, wDefZL, bShL );	{1. neue Leitung einfgen}
	  AddTrmLineAt( aPoint, aPoint1, wDefZL, bShL );	{2. neue Leitung einfgen}
	  END; {WITH aTrmLine}
	END; {IF bPointOnLine}
      END; {IF Objects[ iCount ]}
    Inc( iCount );				{0 ... Count-1}
    END; {WHILE (iCount < Count)}
  END; {IF Count > 0}
Result := bPointOnLine;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.RemoveNodeAt2( thePoint: TPoint ): BOOLEAN;

VAR
  strKey	: ShortString;
  iIndex	: INTEGER;

BEGIN
Result := false;
strKey := BuildNodeCoordinateString( thePoint );
IF Find( strKey, iIndex ) AND (Objects[ iIndex ] IS TNode) THEN
  BEGIN
  Result := DeleteObject( iIndex );
  END; {IF Find( strKey, iIndex )}
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.RemoveNodeAt( aPoint: TPoint ): BOOLEAN;

TYPE
  TLine		= RECORD
		  aTrmLine	: TTrmLine;
		  iIndex	: INTEGER;
		  END;

VAR
  iIndex	: INTEGER;
  iCount	: INTEGER;
  aLine		: TLine;
  aTrmLine	: TTrmLine;
  iFound	: INTEGER;
  strKey	: ShortString;
  bIsStartP	: BOOLEAN;
  bIsStopP	: BOOLEAN;
  checkPoint	: TPoint;

BEGIN
Result	:= false;		{Ergebnis}
strKey	:= BuildNodeCoordinateString( aPoint );
IF Find( strKey, iIndex ) THEN
  BEGIN
  iCount 	:= 0;
  iFound	:= 0;
  WHILE (iCount < Count) AND (iFound < 2) DO
    BEGIN				{nur solange Knoten Endknoten ist!}
    IF Objects[ iCount ] IS TTrmLine THEN
      BEGIN
      aTrmLine	:= TTrmLine( Objects[ iCount ] );
      WITH aTrmLine.Data DO
	BEGIN
	bIsStartP	:= IsSamePoint( tpStart, aPoint );
	bIsStopP	:= IsSamePoint( tpStop, aPoint );;
	IF bIsStartP THEN
	  checkPoint	:= tpStop;
	IF bIsStopP THEN
	  checkPoint	:= tpStart;
	IF (bIsStartP OR bIsStopP) THEN
	  BEGIN
	  aLine.aTrmLine	:= aTrmLine;
	  aLine.iIndex	        := iCount;
	  Inc( iFound );
	  END; {IF (bIsStartP OR bIsStopP)}
	END; {WITH aTrmLine DO}
      END; {IF Objects[ iCount ]}
    Inc( iCount );
    END; {WHILE (iCount < Count)}

  IF (iFound = 1) THEN
    BEGIN					{Endknoten einer Leitung}
    DeleteObject( aLine.iIndex );
    Find( strKey, iIndex );			{zugehrigen Knoten finden: iIndex := iIndex-1!}
    Result	:= DeleteObject( iIndex );
    iCount 	:= 0;			{Gehts am anderen Knoten der gelschte Line weiter?}
    iFound	:= 0;
    WHILE (iCount < Count) AND (iFound < 1) DO
      BEGIN
      IF Objects[ iCount ] IS TTrmLine THEN
	BEGIN
	aTrmLine	:= TTrmLine( Objects[ iCount ] );
	WITH aTrmLine.Data DO
	  BEGIN
	  IF IsSamePoint( tpStart, checkPoint) OR IsSamePoint(tpStop, checkPoint) THEN
	    BEGIN
	    Inc( iFound );
	    END;
	  END; {WITH aTrmLine DO}
	END; {IF Objects[ iCount ]}
      Inc( iCount );
      END; {WHILE (iCount < Count)}

    IF (iFound = 0)  THEN		{gehts weiter?}
      BEGIN				{kann nur ein Knoten sein, Index kann nur "0" sein}
      strKey	:= BuildNodeCoordinateString( checkPoint );
      IF Find( strKey, iIndex ) THEN
	BEGIN
	Result	:= DeleteObject( iIndex );
	END;
      END;

    tMyFileState	:= fstModified;
    END; {IF iFound = 1}
  END; {IF Find( Build}
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.RemoveTrmLine( VAR theTrmLine: TTrmline ): BOOLEAN;

VAR
  iIndex	: INTEGER;
  aTrmLine	: TTrmline;
  iCount	: INTEGER;
  strKey	: ShortString;
  aStartPoint	: TPoint;
  aStopPoint	: TPoint;
  bDelStartNode	: BOOLEAN;
  bDelStopNode	: BOOLEAN;

BEGIN
Result		:= false;						{Ergebnis}
strKey		:= BuildLineCoordinateString( theTrmLine.Data.tpStart, theTrmLine.Data.tpStop );
IF Find( strKey, iIndex ) AND (Objects[ iIndex ] IS TTrmLine) THEN
  BEGIN
  aStartPoint	:= theTrmLine.Data.tpStart;
  aStopPoint	:= theTrmLine.Data.tpStop;

  Result 	:= DeleteObject( iIndex );
  theTrmLine	:= nil;       						{Besser so!}

  bDelStartNode	:= true;
  bDelStopNode	:= true;
  iCount 	:= 0;
  WHILE (iCount < Count) DO
    BEGIN
    IF Objects[ iCount ] IS TTrmLine THEN
      BEGIN
      aTrmLine	:= TTrmLine( Objects[ iCount ] );
      WITH aTrmLine.Data DO
	BEGIN
	IF IsSamePoint( tpStart, aStartPoint ) OR IsSamePoint( tpStop, aStartPoint ) THEN
	  bDelStartNode	:= false;
	IF IsSamePoint( tpStart, aStopPoint ) OR IsSamePoint( tpStop, aStopPoint ) THEN
	  bDelStopNode	:= false;
	END; {WITH aTrmLine}
      END; {IF Objects[ iCount ]}
    Inc( iCount );
    END; {WHILE (iCount < Count)}

  IF bDelStartNode THEN
    RemoveNodeAt2( aStartPoint );

  IF bDelStopNode THEN
    RemoveNodeAt2( aStopPoint );

  END; {IF Find( strKey, iIndex )}
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.ChangeNodeZLAt( aPoint : TPoint; wZL: WORD ): BOOLEAN;

VAR
  iIndex	: INTEGER;
  aNode         : TNode;


BEGIN
IF Find( BuildNodeCoordinateString( aPoint ), iIndex ) THEN
  BEGIN
  Result     := TNode( Objects[ iIndex ] ).Data.wZL <> wZL;
  IF Result THEN
    BEGIN
    aNode	                        := TNode.Create;
    aNode.Data                          := TNode(Objects[iIndex]).Data;
    aNode.Def.Index                     := iIndex;
    aNode.Def.aWa                       := waModify;
    aNode.Def.lTimeStamp                := FTickCount;
    slUndo.AddObject( Strings[iIndex], aNode );
    TNode( Objects[ iIndex ] ).Data.wZL := wZL;
    tMyFileState     := fstModified;
    END;
  END
 ELSE
  BEGIN
  Result	:= false;
  END;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.ChangeNodeSourceAt( aPoint : TPoint; aSource: TSource ): BOOLEAN;

VAR
  iIndex	: INTEGER;
  aNode         : TNode;

BEGIN
IF Find( BuildNodeCoordinateString( aPoint ), iIndex ) THEN
  BEGIN
  Result := TNode( Objects[iIndex] ).Data.theSource <> aSource;
  IF Result THEN
    BEGIN
    aNode	                                    := TNode.Create;
    aNode.Data                                      := TNode(Objects[iIndex]).Data;
    aNode.Def.Index                                 := iIndex;
    aNode.Def.aWa                                   := waModify;
    aNode.Def.lTimeStamp                            := FTickCount;
    slUndo.AddObject( Strings[iIndex], aNode );
    TNode( Objects[ iIndex ] ).Data.theSource       := aSource;
    tMyFileState     := fstModified;
    IF (aSource = srcNone) THEN
      Dec( wSources )
     ELSE
      Inc( wSources );
    END;
  END
 ELSE
  BEGIN
  Result	:= false;
  END;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.ModifyingTrmLine( aTrmLine: TTrmLine ): BOOLEAN;

VAR
  iIndex              : INTEGER;
  newTrmLine          : TTrmLine;

BEGIN
iIndex := IndexOfObject(aTrmLine);
IF (iIndex >= 0) THEN
  BEGIN
  newTrmLine                          := TTrmLine.Create;
  newTrmLine.Data                     := aTrmLine.Data;
  newTrmLine.Def.Index                := iIndex;
  newTrmLine.Def.lTimeStamp           := FTickCount;
  newTrmLine.Def.aWa                  := waModify;
  slUndo.AddObject( Self.Strings[iIndex], newTrmLine );
  tMyFileState                        := fstModified;
  Result                              := true;
  END
 ELSE
  BEGIN
  Result := false;
  END;

END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.ModifyingNode( aNode: TNode ): BOOLEAN;

VAR
  iIndex              : INTEGER;
  newNode             : TNode;

BEGIN
iIndex := IndexOfObject(aNode);
IF (iIndex >= 0) THEN
  BEGIN
  newNode                             := TNode.Create;
  newNode.Data                        := aNode.Data;
  newNode.Def.Index                   := iIndex;
  newNode.Def.lTimeStamp              := FTickCount;
  newNode.Def.aWa                     := waModify;
  slUndo.AddObject( Self.Strings[iIndex], newNode );
  tMyFileState                        := fstModified;
  Result                              := true;
  END
 ELSE
  BEGIN
  Result := false;
  END;

END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.GetNodeAt( aPoint: TPoint ): TNode;

VAR
  iIndex	: INTEGER;

BEGIN
IF Find( BuildNodeCoordinateString( aPoint ), iIndex ) THEN
  GetNodeAt := TNode( Objects[ iIndex ] )
 ELSE
  GetNodeAt := nil;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.GetNodeNameAt( aPoint: TPoint ): ShortString;

VAR
  iIndex	: INTEGER;

BEGIN
IF Find( BuildNodeCoordinateString( aPoint ), iIndex ) THEN
  GetNodeNameAt := Strings[ iIndex ]
 ELSE
  GetNodeNameAt := '';
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.ClearStringList( aStringList: TStringlist ): BOOLEAN;

VAR
  iCount           : INTEGER;
  aNode            : TNode;
  aTrmLine         : TTrmLine;

BEGIN
WITH aStringList DO
  BEGIN
  IF Count > 0 THEN
    BEGIN
    BeginUpdate;
    FOR iCount := 0 TO Count-1 DO
      BEGIN
      IF Objects[ iCount ] IS TNode THEN
        BEGIN
        aNode := TNode( Objects[ iCount ] );
        IF Assigned(aNode) THEN
          aNode.Free
        END
       ELSE
        IF Objects[ iCount ] IS TTrmLine THEN
          BEGIN
          aTrmLine := TTrmLine( Objects[ iCount ] );
          IF Assigned(aTrmLine) THEN
            aTrmLine.Free;
          END; {IF Objects[ iCount ]}
      END; {FOR iCount := 0}
    EndUpdate;
    Clear;
    END; {IF Count > 0}
  END; {WITH aStringList}
Result := true;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.DeleteObject( iIndex: INTEGER ): BOOLEAN;

VAR
  aNode        : TNode;
  aTrmLine     : TTrmLine;

BEGIN
IF (Objects[iIndex] IS TNode) THEN
  BEGIN
  aNode	:= TNode( Objects[iIndex] );
  IF (aNode.Data.theSource = srcPresent) THEN
    Dec( wSources );
  Dec( wNodes );
  aNode.Def.aWA                         := waDelete;
  aNode.Def.Index                       := iIndex;
  aNode.Def.lTimeStamp                  := FTickCount;
  slUndo.AddObject( Strings[iIndex], aNode );
  Delete( iIndex );
  Result                  := true;
  END
 ELSE
  IF (Objects[ iIndex ] IS TTrmLine) THEN
    BEGIN
    aTrmLine                             := TTrmLine( Objects[iIndex] );
    aTrmLine.Def.aWA                     := waDelete;
    aTrmLine.Def.Index                   := iIndex;
    aTrmLine.Def.lTimeStamp              := FTickCount;
    slUndo.AddObject( Strings[iIndex], aTrmLine );
    Delete( iIndex );
    Dec( wLines );
    Result      := true;
    END
   ELSE
    BEGIN
    Result      := false;
    END;

IF Result THEN
  tMyFileState	:= fstModified;
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.UnDo: BOOLEAN;

VAR
  aNode          : TNode;
  aTrmLine       : TTrmLine;
  lTag           : LONGINT;

BEGIN
Result           := false;
WITH slUndo DO
  BEGIN
  IF (Count > 0) THEN
    BEGIN
    lTag := TMyClasses(Objects[Count-1]).Def.lTimeStamp;                           {letzter nderungszeitpunkt}
    WHILE (Count > 0) AND (TMyClasses(Objects[Count-1]).Def.lTimeStamp = lTag) DO
      BEGIN
      IF (Objects[Count-1] IS TNode) AND (TNode(Objects[Count-1]).Def.aWA = waModify) THEN
        BEGIN
        aNode     := TNode( Objects[Count-1] );
        WITH TNode(Self.Objects[aNode.Def.Index]) DO
          BEGIN
          Data        := aNode.Data;                                                 {Daten zurckkopieren}
          Def.aWA     := waRestore;
          END;
        Delete( Count-1 );                                                           {UnDo-Buffer updaten}
        aNode.Free;                                                                  {Object freigeben}
        Result := true;
        END {IF Objects[ Count-1 ] IS TNode}
       ELSE
        IF (Objects[ Count-1 ] IS TTrmLine) AND (TTrmLine(Objects[Count-1]).Def.aWA = waModify) THEN
          BEGIN
          aTrmLine     := TTrmLine( Objects[ Count-1 ] );
          WITH TTrmLine(Self.Objects[aTrmLine.Def.Index]) DO
            BEGIN
            Data        := aTrmLine.Data;                                            {Daten zurckkopieren}
            Def.aWA     := waRestore;
            END;
          Delete( Count-1 );                                                         {UnDo-Buffer updaten}
          aTrmLine.Free;                                                             {Object freigeben}
          Result := true;
          END {IF Objects[ Count-1 ] IS TTrmLine}
         ELSE
          IF (Objects[Count-1] IS TNode) AND (TMyClasses(Objects[Count-1]).Def.aWa = waDelete) THEN
            BEGIN
            aNode := TNode( Objects[Count-1] );
            WITH aNode DO
              BEGIN
              Def.aWA       := waRestore;
              Def.bSelected := false;
              END;
            INHERITED AddObject( Strings[Count-1], Objects[Count-1] );
            Delete( Count-1 );                                                       {UnDo-Buffer updaten}
            Inc(wNodes);
            IF TNode( Objects[Count-1] ).Data.theSource = srcPresent THEN
              Inc(wSources);
            Result      := true;
            END
           ELSE
            IF (Objects[Count-1] IS TTrmline) AND (TMyClasses(Objects[Count-1]).Def.aWa = waDelete) THEN
              BEGIN
              aTrmLine := TTrmline( Objects[Count-1] );
              WITH aTrmLine DO
                BEGIN
		Def.aWA       := waRestore;
		Def.bSelected := false;
		END;
	      INHERITED AddObject( Strings[Count-1], aTrmLine );
	      Delete( Count-1 );                                                     {UnDo-Buffer updaten}
	      Inc(wLines);
	      Result := true;
	      END {IF (Objects[Count-1] IS TTrmline)}
	     ELSE
	      IF (Objects[Count-1] IS TNode) AND (TMyClasses(Objects[Count-1]).Def.aWa = waCreate) THEN
		BEGIN
		aNode := TNode( Objects[Count-1] );
		Dec(wNodes);
		IF aNode.Data.theSource = srcPresent THEN
		  Dec(wSources);
		TNode( Self.Objects[aNode.Def.Index] ).Free;
                Self.Delete( aNode.Def.Index );
                Delete( Count-1 );
                aNode.Free;
                Result      := true;
                END
               ELSE
                IF (Objects[Count-1] IS TTrmline) AND (TMyClasses(Objects[Count-1]).Def.aWa = waCreate) THEN
                  BEGIN
                  aTrmLine := TTrmLine( Objects[Count-1] );
                  Dec(wLines);
                  TTrmLine( Self.Objects[aTrmLine.Def.Index] ).Free;
                  Self.Delete( aTrmLine.Def.Index );
                  Delete( Count-1 );
                  aTrmLine.Free;
                  Result      := true;
                  END {IF (Objects[Count-1] IS TTrmline)}
                 ELSE
                  BEGIN
                  SysUtils.Beep;
                  END;

      END; {WHILE (Count > 0)}
    END; {IF (Count > 0)}
  END; {WITH slUndo}
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.UnDoPossible: BOOLEAN;

BEGIN
Result  := (slUndo.Count > 0);
END;

(*--------------------------------------------------*)
FUNCTION TCanNodes.NetworkExists: BOOLEAN;

BEGIN
Result  := (Count > 0);
END;

(*++++++++++++++++++++++++++++++++++++++++++++++++++*)
END.
