{$I UOpt.INC}				(* Standard Conditional-Defines *)
{.$DEFINE UseSpcBitmap}
{.$DEFine TestWMF}
{.$DEFINE DEMOVERSION}
UNIT SubChild;

INTERFACE

USES
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, ClipBrd,
  UNode, ULine, UTNB,
  CanLines, Menus,
  UOptions;

{$IFDEF DEMOVERSION}
  CONST
    wMaxDemoLines : WORD = 11;
  {$ENDIF}

TYPE
  TMode = ( mNone,
	    mDrawTRMLine,
	    mGetNode, mGetWire,
	    mPlaceZL, mPlaceSrc
	  );

TYPE
  TDefaults	= RECORD
    wNoZL		: WORD;
    wZL			: WORD;
    wLineZL     	: WORD;
    wLength		: WORD;
    bShowNodeNames	: BOOLEAN;
    END;

TYPE
  TPlotMode   = ( pmNone, pmPaint, pmClipboardWMF, pmClipboardEMF, pmWMF, pmEMF );

TYPE
  TSubForm = CLASS(TForm)
    ScrollBox1: TScrollBox;
    PaintBox1: TPaintBox;
    SubFormMenu: TMainMenu;
    Edit1: TMenuItem;
    Undo1: TMenuItem;
    N1: TMenuItem;
    Network1: TMenuItem;
    Options1: TMenuItem;
    PROCEDURE PaintBox1Paint(Sender: TObject);
    PROCEDURE PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    PROCEDURE PaintBox1MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    PROCEDURE PaintBox1MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    PROCEDURE FormDestroy(Sender: TObject);
    PROCEDURE FormActivate(Sender: TObject);
    PROCEDURE FormDeactivate(Sender: TObject);
    PROCEDURE Network1Click(Sender: TObject);
    PROCEDURE Undo1Click(Sender: TObject);
    PROCEDURE Options1Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
  PRIVATE
    theNodes		: TCanNodes;
    theParentPaintBox	: TPaintBox;
    theParentWindow     : THandle;
    theStartPoint	: TPoint;
    theStopPoint  	: TPoint;
    theSnappiness 	: TPoint;
    theMultiplier       : TPoint;
    theDefaults		: TDefaults;
    theMode		: TMode;
    wNodeRadius		: WORD;
    bMouseCaptured	: BOOLEAN;
    PROCEDURE UpdateMenuItems(Sender: TObject);
    PROCEDURE SelectMode( aMode: TMode );
    PROCEDURE DisplayMouseCoordinates( aPoint: TPoint );
    PROCEDURE SnapToGrid( VAR aPoint: TPoint );
    PROCEDURE SnapToGridLine( VAR aPoint: TPoint );
    PROCEDURE DrawOutlineOnly( aCanvas : TCanvas; iX, iY : INTEGER );

    FUNCTION  DrawNodeAt( aCanvas: TCanvas; aPoint: TPoint; bSelected: BOOLEAN; aSizeLog: TPoint ): BOOLEAN;
    FUNCTION  DrawNode( aCanvas: TCanvas; hFontHorz: THandle; aNode: TNode; aSizeLog: TPoint ): BOOLEAN;

    FUNCTION  DrawZLAndSrc( aCanvas: TCanvas; aNode: TNode; aSizeLog: TPoint ): BOOLEAN;
    FUNCTION  DrawLine( aCanvas: TCanvas; aTrmLine: TTrmLine; aSizeLog: TPoint; aPlotMode: TPlotMode ): BOOLEAN;
    PROCEDURE SpeedButtonsVisible( bVisible: BOOLEAN );
    FUNCTION  CoCo( aPoint: TPoint ): TPoint;

    PROCEDURE SetScrollBarRanges( aPoint: TPoint );
    FUNCTION  GetScrollBarRanges: TPoint;
    FUNCTION  GetScrollBarPositions: TPoint;

    FUNCTION  SetMappingMode( aCanvas: TCanvas ): BOOLEAN;
    FUNCTION  DrawNodeImmediately( aNode: TNode ): BOOLEAN;
    FUNCTION  DrawLineImmediately( aTrmLine: TTrmLine ): BOOLEAN;

  PUBLIC
    CONSTRUCTOR CreateExtra( AOwner: TComponent;
                             aParentPaintBox: TPaintBox;
                             aFileName: TFileName;
                             aSizeLog: TPoint;
                             aMultiplier: TPoint;
                             aParentWnd: THandle  );

    procedure WMNCLButtonDown(var Message: TWMNCLButtonDown); message WM_NCLBUTTONDOWN;

  PUBLISHED
    FUNCTION  UnDo: BOOLEAN;
    PROCEDURE Network( aOwner: TComponent );
    FUNCTION  SaveCANAs( aOwner: TComponent ): TFileName;
    FUNCTION  SaveCAN: TFileName;
    FUNCTION  SaveCircuit( aOwner: TComponent ): TFileName;
    PROCEDURE PlotIt( aCanvas: TCanvas; aPlotMode: TPlotMode; bShowText: BOOLEAN; aSizeLog: TPoint );
    PROPERTY  FMode: TMode READ theMode WRITE SelectMode;
    PROPERTY  FDimensions: TPoint READ GetScrollBarRanges WRITE SetScrollBarRanges;
    PROPERTY  FPositions: TPoint READ GetScrollBarPositions;
    END;


IMPLEMENTATION

USES
  Main, ChildWin, Math;

{$R *.DFM}

(*----------------------------------------------------*)
FUNCTION IMin( i1, i2: INTEGER ): INTEGER;

BEGIN
IF i1 < i2 THEN
  Result := i1
 ELSE
  Result := i2;
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.WMNCLButtonDown(var Message: TWMNCLButtonDown);

BEGIN
IF WindowState = wsMinimized THEN
  Message.HitTest := HTNoWhere;
INHERITED;
END;

(*------------------------------------------------------*)
FUNCTION TSubForm.DrawNodeAt( aCanvas: TCanvas; aPoint: TPoint; bSelected: BOOLEAN; aSizeLog: TPoint ): BOOLEAN;

VAR
  aColor	: TColor;

BEGIN
WITH aCanvas DO
  BEGIN
  aColor := Brush.Color;
  IF bSelected THEN
    Brush.Color := clRed
   ELSE
    Brush.Color := clTeal;
  WITH aPoint DO
    Ellipse( (X-wNodeRadius)*aSizeLog.X, (Y-wNodeRadius)*aSizeLog.Y,
	     (X+wNodeRadius)*aSizeLog.X, (Y+wNodeRadius)*aSizeLog.Y );
  Brush.Color := aColor;
  END;
Result := true;
END;

(*------------------------------------------------------*)
FUNCTION TSubForm.DrawZLAndSrc( aCanvas: TCanvas; aNode: TNode; aSizeLog: TPoint ): BOOLEAN;

VAR
  apResi        : TPoint;
  apSubNode     : TPoint;
  apGnd         : TPoint;
  aBrushColor	: TColor;
  aPenStyle     : TPenStyle;

BEGIN
WITH aCanvas, aNode.Data.thePoint DO
  BEGIN
  IF (aNode.Data.wZL <> theNodes.FNoZL) OR (aNode.Data.theSource = srcPresent) THEN
    BEGIN
    apResi.X    := theSnappiness.X   DIV 16;
    apResi.Y    := theSnappiness.Y   DIV  4;
    apSubNode.X := theSnappiness.X   DIV  6;
    apSubNode.Y := theSnappiness.Y   DIV  4;
    apGnd.X     := apSubNode.X;
    apGnd.Y     := 3*apResi.Y;
    MoveTo( X*aSizeLog.X, Y*aSizeLog.Y );              		 	                {Knoten}
    LineTo( (X+apSubNode.X)*aSizeLog.X, (Y-apSubNode.Y)*aSizeLog.Y );                   {Sub-knoten}
    LineTo( (X+apSubNode.X)*aSizeLog.X, (Y - apGnd.Y)*aSizeLog.Y );                     {zum GND}
    MoveTo( (X+apSubNode.X - apResi.X)*aSizeLog.X, (Y - apGnd.Y)*aSizeLog.Y );          {GND, links}
    LineTo( (X+apSubNode.X + apResi.X)*aSizeLog.X, (Y - apGnd.Y)*aSizeLog.Y );          {GND, rechts}

    aBrushColor := Brush.Color;
    aPenStyle   := Pen.Style;
    IF (aNode.Data.wZL <> theNodes.FNoZL) THEN
      BEGIN
      Brush.Color	:= clYellow;
      RectAngle( (X+apSubNode.X - apResi.X)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - apResi.Y )*aSizeLog.Y,
                 (X+apSubNode.X + apResi.X)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - 2*apResi.Y  )*aSizeLog.Y );

      IF (aNode.Data.theSource = srcPresent) THEN
        BEGIN
        Brush.Color	:= clRed;
        Pen.Style       := psClear;
	Polygon( [Point((X+apSubNode.X)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - apResi.Y )*aSizeLog.Y),
                  Point((X+apSubNode.X - apResi.X+1)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - 2*apResi.Y +1 )*aSizeLog.Y ),
                  Point((X+apSubNode.X + apResi.X-1)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - 2*apResi.Y +1 )*aSizeLog.Y )
                 ] );
        END; {IF (aNode.theSource = srcPresent)}
      END
     ELSE
      IF (aNode.Data.wZL = theNodes.FNoZL) AND (aNode.Data.theSource = srcPresent) THEN
        BEGIN
        Brush.Color	:= clWhite;
        RectAngle( (X+apSubNode.X - apResi.X)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - apResi.Y )*aSizeLog.Y,
                   (X+apSubNode.X + apResi.X)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - 2*apResi.Y  )*aSizeLog.Y );

        Brush.Color     := clRed;
        Pen.Style       := psClear;
        Polygon( [Point((X+apSubNode.X)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - apResi.Y )*aSizeLog.Y),
                  Point((X+apSubNode.X - apResi.X+1)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - 2*apResi.Y +1 )*aSizeLog.Y ),
                  Point((X+apSubNode.X + apResi.X-1)*aSizeLog.X, (Y - apSubNode.Y DIV 2 - 2*apResi.Y +1 )*aSizeLog.Y )
                 ] );
        END; {IF (aNode.wZL = theNodes.FNoZL)}
    Brush.Color := aBrushColor;
    Pen.Style   := aPenStyle;
    END; {IF (aNode.wZL <> theNodes.FNoZL) OR ...}
  END; {WITH aCanvas}
Result := true;
END;

(*------------------------------------------------------*)
FUNCTION TSubForm.DrawNode( aCanvas: TCanvas; hFontHorz: THandle; aNode: TNode; aSizeLog: TPoint ): BOOLEAN;

VAR
  iOldBkMode    : INTEGER;
  hOldFont      : HFont;

BEGIN
DrawNodeAt( aCanvas, aNode.Data.thePoint, aNode.Def.bSelected, aSizeLog );
WITH aCanvas DO
  BEGIN
  IF aNode.Data.bShowName AND theDefaults.bShowNodeNames AND (hFontHorz <> 0) THEN
    BEGIN
    WITH aNode.Data DO
      BEGIN
      IF (hFontHorz <> 0) THEN
	BEGIN
        iOldBkMode := SetBkMode( Handle, transparent );
        hOldFont := SelectObject( Handle, hFontHorz );
	SetTextAlign( Handle, ta_Right OR ta_Top );
	TextOut( (thePoint.X - wNodeRadius)*aSizeLog.X, (thePoint.Y-2)*aSizeLog.Y, strName );
        SelectObject( Handle, hOldFont );
        SetBkMode( Handle, iOldBkMode );
	END;
      END;
    END;
  END;
Result := true;
END;

(*------------------------------------------------------*)
FUNCTION TSubForm.DrawNodeImmediately( aNode: TNode ): BOOLEAN;

BEGIN
SetMappingMode( PaintBox1.Canvas );
Result := DrawNode( PaintBox1.Canvas, 0, aNode, Point(1,1) );
END;

(*------------------------------------------------------*)
FUNCTION TSubForm.DrawLine( aCanvas: TCanvas; aTrmLine: TTrmLine; aSizeLog: TPoint; aPlotMode: TPlotMode ): BOOLEAN;

VAR
  iXCenter	: INTEGER;
  iYCenter	: INTEGER;
  hOldFont	: THandle;
  aPenColor	: TColor;
  aFontColor	: TColor;
  hNewFont      : THandle;
  aLogFont      : TLogFont;
  iOldBkMode    : INTEGER;

BEGIN
WITH aCanvas DO
  BEGIN
  aFontColor	:= Font.Color;
  aPenColor	:= Pen.Color;
  IF aTrmLine.Def.bSelected THEN
    Pen.Color	:= clRed
   ELSE
    Pen.Color	:= clBlack;
  WITH aTrmLine.Data.tpStart DO MoveTo( X*aSizeLog.X, Y*aSizeLog.Y );
  WITH aTrmLine.Data.tpStop DO LineTo( X*aSizeLog.X, Y*aSizeLog.Y );
  IF aTrmLine.Data.bShowLength THEN
    BEGIN
    WITH aTrmLine.Data, aLogFont DO
      BEGIN
      lfWidth		:= 0;
      IF (fsBold IN Font.Style) THEN lfWeight := fw_Bold ELSE lfWeight := fw_Normal;
      lfItalic          := BYTE(fsBold IN Font.Style);
      lfUnderline       := BYTE(fsunderline IN Font.Style);
      lfStrikeOut      	:= BYTE(fsStrikeOut IN Font.Style);
      lfClipPrecision 	:= Clip_Character_Precis OR Clip_LH_Angles;
      lfCharSet		:= ANSI_Charset;
      lfOutPrecision  	:= out_tt_only_precis;
      lfQuality		:= Proof_Quality;
      lfPitchAndFamily	:= Variable_Pitch;
      StrPCopy( lfFaceName, aCanvas.Font.Name );
      CASE aPlotMode OF
        pmPaint:
          BEGIN
          lfEscapement      := Round(10.0*RadToDeg(-sAngle));
          lfOrientation	    := lfEscapement;
          lfHeight          := -MulDiv( aCanvas.Font.Size, aCanvas.Font.PixelsPerInch, 72 );
          END;
        pmClipBoardWMF, pmWMF:
          BEGIN
          lfEscapement      := Round(10.0*RadToDeg(sAngle));
          lfOrientation	    := lfEscapement;
          lfHeight          := -MulDiv( aCanvas.Font.Size, 2540, 72 );
          END;
        pmClipBoardEMF, pmEMF:
          BEGIN
          lfEscapement      := Round(10.0*RadToDeg(-sAngle));
          lfOrientation	    := lfEscapement;
          lfHeight          := -MulDiv( aCanvas.Font.Size, 2540, 72 );
          END;
        END; {CASE aPlotMode}

      hNewFont := CreateFontIndirect( aLogFont );
      IF (wLength = 0) THEN
        Font.Color := clRed;
      iXCenter     := tpStart.X+(tpStop.X-tpStart.X) DIV 2;
      iYCenter     := tpStart.Y+(tpStop.Y-tpStart.Y) DIV 2;
      hOldFont     := SelectObject( Handle, hNewFont );
      iOldBkMode   := SetBkMode( Handle, transparent );
      SetTextAlign( Handle, ta_Bottom OR ta_Center );
      TextOut( iXCenter*aSizeLog.X, iYCenter*aSizeLog.Y, IntToStr(wLength)+'mm' );
      SetBkMode( Handle, iOldBkMode );
      SelectObject( Handle, hOldFont );
      DeleteObject( hNewFont );
      END; {WITH aTrmLine DO}
    END; {IF aTrmLine.bShowLength AND}
  Pen.Color	:= aPenColor;
  Font.Color	:= aFontColor;
  END; {WITH aCanvas DO}
Result := true;
END;

(*------------------------------------------------------*)
FUNCTION TSubForm.DrawLineImmediately( aTrmLine: TTrmLine ): BOOLEAN;

BEGIN
SetMappingMode( PaintBox1.Canvas );
Result := DrawLine( PaintBox1.Canvas, aTrmLine, Point(1,1), pmPaint );
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.PlotIt( aCanvas: TCanvas; aPlotMode: TPlotMode; bShowText: BOOLEAN; aSizeLog: TPoint );

VAR
  aTrmLine	: TTrmLine;
  aNode		: TNode;
  iCount	: INTEGER;
  aLogFont	: TLogFont;
  hFontHorz	: THandle;
  hOldFont	: THandle;

BEGIN
WITH aCanvas DO
  BEGIN
  {$IFDEF TestWMF}
    Brush.Color := clYellow;
    Pen.Color   := clBlack;
    Ellipse( 0, 0, 1000, 1000 );
    Brush.Color := clRed;
    Ellipse( 0, 0, 500, 500 );
    Brush.Color := clBlue;
    Ellipse( 0, 0, 100, 100 );
    Brush.Color := clAqua;
    Ellipse( 0, 0, 60, 60 );
  {$ELSE}
    IF bShowText THEN
      BEGIN
      hOldFont	:= aCanvas.Font.Handle;
      WITH aLogFont DO
	BEGIN
	CASE aPlotMode OF
	  pmPaint:
	    BEGIN
	    Font     := PaintBox1.Font;
	    lfHeight := -MulDiv( aCanvas.Font.Size, aCanvas.Font.PixelsPerInch, 72 );
	    END;
	  pmClipBoardWMF, pmWMF,
	  pmClipBoardEMF, pmEMF:
	    BEGIN
	    Font     := theParentPaintbox.Font;
	    lfHeight := -MulDiv( aCanvas.Font.Size, 2540, 72 );
	    END;
	  END; {CASE aPlotMode}
	lfWidth			:= 0;
	lfEscapement		:= 0;
	lfOrientation	        := lfEscapement;
	IF (fsBold IN Font.Style) THEN lfWeight := fw_Bold ELSE lfWeight := fw_Normal;
	lfItalic                := BYTE(fsBold IN Font.Style);
	lfUnderline             := BYTE(fsunderline IN Font.Style);
	lfStrikeOut      	:= BYTE(fsStrikeOut IN Font.Style);
	lfClipPrecision		:= Clip_Character_Precis OR Clip_LH_Angles;
	lfCharSet		:= ANSI_Charset;
	lfOutPrecision		:= out_tt_only_precis;
	lfQuality		:= Proof_Quality;
	lfPitchAndFamily	:= Variable_Pitch;
	StrPCopy( lfFaceName, aCanvas.Font.Name );
	END; {WITH aLogFont}
      hFontHorz := CreateFontIndirect( aLogFont );
      END {IF bShowText}
     ELSE
      BEGIN
      hFontHorz := NULL;
      END; {IF bShowText}

    WITH theNodes DO
      BEGIN
      EnumerateNodes;								// Nummerierung basteln
      IF Count > 0 THEN
	BEGIN
	FOR iCount := 0 TO Count-1 DO
	  BEGIN
	  IF Objects[ iCount ] IS TTrmLine THEN
	    BEGIN
	    aTrmLine := TTrmLine( Objects[ iCount ] );
	    DrawLine( aCanvas, aTrmLine, aSizeLog, aPlotMode );
	    END;
	  END; {FOR iCount := 0}
	FOR iCount := 0 TO Count-1 DO
	  BEGIN
	  IF Objects[ iCount ] IS TNode THEN
	    BEGIN
	    aNode := TNode( Objects[ iCount ] );
	    DrawZLAndSrc( aCanvas, aNode, aSizeLog );
	    DrawNode( aCanvas, hFontHorz, aNode, aSizeLog );
	    END;
	  END; {FOR iCount := 0}
	END; {IF Count > 0}
      END; {WITH theNodes}

    IF (hFontHorz <> NULL) THEN
      BEGIN
      SelectObject( aCanvas.Handle, hOldFont );
      DeleteObject( hFontHorz );
      END;
    {$ENDIF}
  END;
END;

(*----------------------------------------------------*)
FUNCTION  TSubForm.SetMappingMode( aCanvas: TCanvas ): BOOLEAN;

BEGIN
WITH aCanvas DO
  BEGIN
  SetMapMode( Handle, mm_IsoTropic );
  SetWindowExtEx( Handle,
		  IMin( PaintBox1.ClientWidth, FDimensions.X ),
		  IMin( PaintBox1.ClientHeight, FDimensions.Y ),
		  nil
		);
  SetViewPortExtEx( Handle, FDimensions.X, -FDimensions.Y, nil );
  SetWindowOrgEx( Handle,
		  0,
		  IMin( PaintBox1.ClientHeight, FDimensions.Y ),
		  nil
		);
  SetViewPortOrgEx( Handle, -FPositions.X, -FPositions.Y, nil );
  END;
Result := true;
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.PaintBox1Paint(Sender: TObject);

VAR
  aCanvas	: TCanvas;
  {$IFDEF UseSpcBitmap}
    aBitMap	: TBitMap;
    {$ENDIF}

BEGIN
{$IFDEF UseSpcBitmap}
  aBitMap	:= TBitMap.Create;
  WITH aBitMap DO
    BEGIN
    Height	:= IMin( PaintBox1.ClientHeight, FDimensions.Y );
    Width	:= IMin( PaintBox1.ClientWidth, FDimensions.X );
    END; {WITH aBitMap}
  aCanvas	:= aBitMap.Canvas;
  aCanvas.Font	:= PaintBox1.Canvas.Font;
 {$ELSE}
  aCanvas	:= PaintBox1.Canvas;
  {$ENDIF}

SetMappingMode( aCanvas );
WITH aCanvas DO
  BEGIN
  Pen.Style := psDot;
  MoveTo( 0, 0 );
  LineTo( IMin(PaintBox1.ClientWidth, FDimensions.X) , 0 );
  LineTo( IMin(PaintBox1.ClientWidth, FDimensions.X), IMin(PaintBox1.ClientHeight, FDimensions.Y) );
  LineTo( 0, IMin(PaintBox1.ClientHeight, FDimensions.Y) );
  LineTo( 0, 0 );
  Pen.Style := psSolid;
  PlotIt( aCanvas, pmPaint, true, Point(1,1) );
  END; {WITH aCanvas}

{$IFDEF UseSpcBitmap}
  WITH aBitMap DO
    BEGIN
    BitBlt( PaintBox1.Canvas.Handle, 0, 0, Width, Height, aBitMap.Canvas.Handle, 0, 0, SRCand );
    Free;
    END; {WITH aBitMap}
  {$ENDIF}
END;

(*----------------------------------------------------*)
FUNCTION TSubForm.CoCo( aPoint: TPoint ): TPoint;

BEGIN
WITH Result DO
  BEGIN
  X := aPoint.X;
  Y := FDimensions.Y-aPoint.Y;
  END;
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.DisplayMouseCoordinates( aPoint: TPoint );

BEGIN
WITH TMDIMainForm(Application.Mainform).StatusBar DO
  BEGIN
  Panels[1].Text := 'X:'+IntToStr( aPoint.X );
  Panels[2].Text := 'Y:'+IntToStr( FDimensions.Y-aPoint.Y );
  END;
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.SelectMode( aMode: TMode );

BEGIN
theMode := aMode;
END;

(*------------------------------------------------------*)
PROCEDURE TSubForm.SnapToGrid( VAR aPoint: TPoint );

BEGIN
WITH aPoint DO
  BEGIN
  X := (
	     ( ( X + theSnappiness.X SHR 1) DIV theSnappiness.X ) * theSnappiness.X
       );
  Y := (
	     ( ( Y + theSnappiness.Y SHR 1) DIV theSnappiness.Y ) * theSnappiness.Y
       );
  END;

END;

(*------------------------------------------------------*)
PROCEDURE TSubForm.SnapToGridLine( VAR aPoint: TPoint );

VAR
  theSnappinessLine : TPoint;

BEGIN
theSnappinessLine.X := theSnappiness.X SHR 1;
theSnappinessLine.Y := theSnappiness.Y SHR 1;
WITH aPoint DO
  BEGIN
  X := (
	     ( ( X + theSnappinessLine.X SHR 1) DIV theSnappinessLine.X ) * theSnappinessLine.X
       );
  Y := (
	     ( ( Y + theSnappinessLine.Y SHR 1) DIV theSnappinessLine.Y ) * theSnappinessLine.Y
       );
  END;

END;

(*----------------------------------------------------*)
FUNCTION TSubForm.GetScrollBarRanges: TPoint;

BEGIN
WITH Result, ScrollBox1 DO
  BEGIN
  X := HorzScrollBar.Range;
  Y := VertScrollBar.Range;
  END;
END;
(*----------------------------------------------------*)
PROCEDURE TSubForm.SetScrollBarRanges( aPoint: TPoint );

BEGIN
WITH ScrollBox1 DO
  BEGIN
  HorzScrollBar.Range	:= aPoint.X;
  VertScrollBar.Range	:= aPoint.Y;
  IF VertScrollBar.Position > 0 THEN			// wg. Fehler in Bildwiederherstellung
    VertScrollBar.Position := VertScrollBar.Position-1
   ELSE
    VertScrollBar.Position := VertScrollBar.Position+1;
  END;
END;

(*----------------------------------------------------*)
FUNCTION TSubForm.GetScrollBarPositions: TPoint;

BEGIN
WITH Result, ScrollBox1 DO
  BEGIN
  X := ScrollBox1.HorzScrollBar.Position;
  Y := ScrollBox1.VertScrollBar.Position;
  END;
END;

(*----------------------------------------------------*)
CONSTRUCTOR TSubForm.CreateExtra( AOwner: TComponent;
				  aParentPaintBox: TPaintBox;
				  aFileName: TFileName;
				  aSizeLog: TPoint;
				  aMultiplier: TPoint;
				  aParentWnd: THandle );

BEGIN
INHERITED Create( AOwner );
theMultiplier           := aMultiplier;
theMode		        := mNone;
theParentWindow         := aParentWnd;
wNodeRadius		:= 5;
theParentPaintBox	:= aParentPaintBox;
WITH theSnappiness DO
  BEGIN
  X	:= 100;
  Y	:= 100;
  END;

WITH theDefaults DO
  BEGIN
  wNoZL		:= 1500;
  wZL		:= 120;
  wLineZL       := 120;
  wLength	:= 2000;
  bShowNodeNames:= true;
  theNodes	:= TCanNodes.CreateCANLines( wZL, wNoZL, wLineZL );
  theNodes.Load( aFileName );
  Caption       := theNodes.FFileName;
  END;

IF (theNodes.FNumNodes > 0) THEN
  BEGIN
  FDimensions := Point(
		       theNodes.FMax.X + 2*theSnappiness.X - theNodes.FMin.X,
		       theNodes.FMax.Y + 2*theSnappiness.Y - theNodes.FMin.Y
		      );
  theNodes.ShiftNetwork( Point(-theNodes.FMin.X+theSnappiness.X, -theNodes.FMin.Y+theSnappiness.Y) );
  SendMessage( theParentWindow, CM_ParentCalled, cmPCChangeSizeX, FDimensions.X );
  SendMessage( theParentWindow, CM_ParentCalled, cmPCChangeSizeY, FDimensions.Y );
  END
 ELSE
  BEGIN
  FDimensions := aSizeLog;
  END;
END;

(*------------------------------------------------------*)
PROCEDURE TSubForm.DrawOutlineOnly( aCanvas: TCanvas; iX, iY : INTEGER);

BEGIN
WITH aCanvas DO
  BEGIN
  SetROP2( Handle, R2_NOTXORPEN );
  Pen.Style := psDot;
  Brush.Style := bsClear;
  MoveTo( theStartPoint.X,theStartPoint.Y );
  LineTo( theStopPoint.X,theStopPoint.Y );
  theStopPoint.X := iX;
  theStopPoint.Y := iY;
  SnapToGrid( theStopPoint );
  MoveTo( theStartPoint.X, theStartPoint.Y );
  LineTo( theStopPoint.X, theStopPoint.Y );
  END;
END;

(*------------------------------------------------------*)
PROCEDURE TSubForm.PaintBox1MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

BEGIN
CASE theMode OF
 mDrawTRMLine:
  BEGIN
  bMouseCaptured	:= SetCapture( GetCapture ) <> 0;
  theStartPoint.X	:= X;
  theStartPoint.Y	:= Y;
  SnapToGrid( theStartPoint );
  theStopPoint 		:= theStartPoint;
  PaintBox1.Canvas.MoveTo( X, Y );
  END
 ELSE
  END; {CASE theMode OF};
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);

VAR
  aPoint : TPoint;

BEGIN
DisplayMouseCoordinates( Point( X, Y ) );
IF bMouseCaptured THEN
  BEGIN
  DrawOutlineOnly( PaintBox1.Canvas, X, Y )
  END
 ELSE
  BEGIN
  aPoint.X := X;
  aPoint.Y := FDimensions.Y-Y;
  SnapToGrid( aPoint );
  END;
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.PaintBox1MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);


VAR
  aPoint	: TPoint;
  aNode		: TNode;
  aTrmLine	: TTrmline;
  iCode		: INTEGER;
  strDummy	: ShortString;

BEGIN
theNodes.FTickCount := GetTickCount;
IF bMouseCaptured THEN
  BEGIN
  DrawOutlineOnly( Canvas, X, Y );
  WITH PaintBox1.Canvas DO
    BEGIN
    SetROP2( Handle, R2_COPYPEN );
    Pen.Style := psSolid;
    SnapToGrid( theStopPoint );
    IF (theStartPoint.X <> theStopPoint.X) OR (theStartPoint.Y <> theStopPoint.Y) THEN
      BEGIN
      {$IFNDEF DEMOVERSION}
	theNodes.AddNodeAt( CoCo( theStartPoint ) );
	theNodes.AddNodeAt( CoCo( theStopPoint ) );
	theNodes.AddTrmLineAt( CoCo( theStartPoint ), CoCo( theStopPoint ), theDefaults.wLineZL, true );
       {$ELSE}
	IF theNodes.FNumLines < wMaxDemoLines THEN
	  BEGIN
	  theNodes.AddNodeAt( CoCo( theStartPoint ) );
	  theNodes.AddNodeAt( CoCo( theStopPoint ) );
	  theNodes.AddTrmLineAt( CoCo( theStartPoint ), CoCo( theStopPoint ), theDefaults.wLineZL, true );
	  END
	 ELSE
	  BEGIN
	  TMDIMainForm(Application.Mainform).SpeedButtonLTG.Visible 	:= false;
	  theMode                                                       := mNone;
	  MessageDlg( 'DEMO: maximum number of CAN lines'#10#13' exceeded!',
		       mtInformation, [mbOK], 0);
	  END;
	{$ENDIF}

      IF (theNodes.FMin.Y <= 0) OR (theNodes.FMax.X >= FDimensions.X) OR
	 (theNodes.FMax.Y >= FDimensions.Y) OR (theNodes.FMin.X <= 0) THEN
	BEGIN
	theNodes.ShiftNetwork( Point(-theNodes.FMin.X+theSnappiness.X, -theNodes.FMin.Y+theSnappiness.Y) );
	FDimensions := Point(
			     theNodes.FMax.X + 2*theSnappiness.X - theNodes.FMin.X,
			     theNodes.FMax.Y + 2*theSnappiness.Y - theNodes.FMin.Y
			    );
	SendMessage( theParentWindow, CM_ParentCalled, cmPCChangeSizeX, FDimensions.X );
	SendMessage( theParentWindow, CM_ParentCalled, cmPCChangeSizeY, FDimensions.Y );
	END;
      theParentPaintBox.RePaint;
      PaintBox1.Repaint;
      END;
    END;
  ReleaseCapture;
  bMouseCaptured := FALSE;
  UpdateMenuItems( Self );
  Exit;					{hier gehts raus}
  END;

CASE theMode OF
 mPlaceZL:
  BEGIN
  aPoint.X := X;
  aPoint.Y := Y;
  SnapToGrid( aPoint );
  IF theNodes.NodeExistAt( CoCo( aPoint ) ) THEN
    BEGIN
    aNode := theNodes.GetNodeAt( CoCo( aPoint ) );
    IF theNodes.ModifyingNode( aNode ) THEN
      BEGIN
      aNode.Data.wZL := theDefaults.wZL;
      PaintBox1.Repaint;
      theParentPaintBox.Repaint;
      END;
    END;
  END; {mPlaceZL:}
 mPlaceSrc:
  BEGIN
  aPoint.X := X;
  aPoint.Y := Y;
  SnapToGrid( aPoint );
  IF theNodes.NodeExistAt( CoCo( aPoint ) ) THEN
    BEGIN
    aNode := theNodes.GetNodeAt( CoCo( aPoint ) );
    IF theNodes.ModifyingNode( aNode ) AND (aNode.Data.theSource = srcNone) THEN
      BEGIN
      aNode.Data.theSource := srcPresent;
      PaintBox1.Repaint;
      theParentPaintBox.Repaint;
      END;
    {$IFDEF JustOneSource}
      IF theNodes.FNumSources > 0 THEN
	BEGIN
	tbSource.Visible	:= false;
	theMode			:= mNone;
	END;
      {$ENDIF}
    END;
  END; {mPlaceSrc:}
 mGetNode:
  BEGIN
  aPoint.X := X;
  aPoint.Y := Y;
  SnapToGrid( aPoint );
  IF theNodes.NodeExistAt( CoCo( aPoint ) ) THEN
    BEGIN
    FKnoten := TFKnoten.Create( Application );
    FKnoten.FDefaultZL	:= theDefaults.wZL;
    FKnoten.FDefaultNoZL:= theNodes.FNoZL;
    WITH FKnoten DO
      BEGIN
      Left			:= PaintBox1.ClientToScreen(aPoint).X;
      Top 			:= PaintBox1.ClientToScreen(aPoint).Y;
      aNode 			:= theNodes.GetNodeAt( CoCo( aPoint ) );
      aNode.Def.bSelected	:= true;
      DrawNodeImmediately( aNode );
      CheckBoxQuelle.Checked	:= aNode.Data.theSource = srcPresent;
      CheckBoxName.Checked	:= aNode.Data.bShowName;
      EditName.Text 		:= aNode.Data.strName;
      EditZL.Text 		:= IntToStr( aNode.Data.wZL );

      {$IFDEF JustOneSource}
	IF (theNodes.FNumSources > 0) AND (aNode.theSource = srcNone) THEN
	  BEGIN
	  CheckBoxQuelle.Visible	:= false;
	  END
	 ELSE
	  BEGIN
	  CheckBoxQuelle.Visible	:= true;
	  END;
	{$ENDIF}

      END;
    CASE FKnoten.ShowModal OF
      mrOK:
	BEGIN
	WITH FKnoten DO
	  BEGIN
	  aNode.Def.bSelected	:= false;
	  theNodes.ModifyingNode( aNode );
	  IF CheckBoxQuelle.Checked THEN
	    aNode.Data.theSource := srcPresent
	   ELSE
	    aNode.Data.theSource := srcNone;
	  aNode.Data.strName 	:= EditName.Text;
	  aNode.Data.bShowName	:= CheckBoxName.Checked;
	  strDummy		:= EditZL.Text;
	  WHILE Pos( ' ', strDummy) <> 0 DO
	    Delete( strDummy, Pos( ' ', strDummy), 1 );
	  aNode.Data.wZL      := StrToInt( strDummy );
	  anode.Def.aWA       := waModify;
	  END;
	{$IFDEF JustOneSource}
	  tbSource.Visible	:= (theNodes.FNumSources = 0);
	  {$ENDIF}
	PaintBox1.Refresh;
	theParentPaintBox.Refresh;
	END;
      mrCancel:
	BEGIN
	aNode.Def.bSelected	:= false;
	DrawNodeImmediately( aNode );
	END;
      mrAbort:
	BEGIN
	IF theNodes.RemoveNodeAt( CoCo( aPoint ) ) THEN
	  BEGIN
	  PaintBox1.Refresh;
	  theParentPaintBox.Refresh;
	  END
	 ELSE
	  BEGIN
	  aNode.Def.bSelected	:= false;
	  DrawNodeImmediately( aNode );
	  END;
	END;
      END;
    {$IFDEF JustOneSource}
      tbSource.Visible	:= (theNodes.FNumSources < 1);
      {$ENDIF}
    FKnoten.Free;
    END;
  END; {mGetNode:}

 mGetWire:
  BEGIN
  aPoint.X	:= X;
  aPoint.Y	:= Y;
  aTrmLine	:= theNodes.GetTrmLineAt( CoCo( aPoint ) );
  IF aTrmLine <> nil THEN
    BEGIN
    aTrmLine.Def.bSelected := true;
    DrawLineImmediately( aTrmLine );
    LineDlg			:= TLineDlg.Create( Application );
    WITH LineDlg DO
      BEGIN
      Left		        := PaintBox1.ClientToScreen(aPoint).X;		{!!!}
      Top 		        := PaintBox1.ClientToScreen(aPoint).Y;
      FDefaultZL		:= theDefaults.wZL;
      FDefaultLength		:= theDefaults.wLength;
      EditZL.Text		:= IntToStr( aTrmLine.Data.wZl );
      EditLength.Text		:= IntToStr( aTrmLine.Data.wLength );
      CheckBoxShowLength.Checked:= aTrmLine.Data.bShowLength;
      END;
    CASE LineDlg.ShowModal OF
      mrOK:
	BEGIN
	aTrmLine.Def.bSelected := false;
	theNodes.ModifyingTrmLine( aTrmLine );
	strDummy		        := LineDlg.EditZL.Text;
	WHILE Pos( ' ', strDummy) <> 0 DO Delete( strDummy, Pos( ' ', strDummy), 1 );
	Val( strDummy, aTrmLine.Data.wZl, iCode );
	strDummy		        := LineDlg.EditLength.Text;
	WHILE Pos( ' ', strDummy) <> 0 DO Delete( strDummy, Pos( ' ', strDummy), 1 );
	Val( strDummy, aTrmLine.Data.wLength, iCode );
	aTrmLine.Data.bShowLength	:= LineDlg.CheckBoxShowLength.Checked;

	PaintBox1.Refresh;
	theParentPaintBox.Refresh;
	END;
      mrAbort:
	BEGIN
	IF theNodes.RemoveTrmLine( aTrmLine ) THEN
	  BEGIN
	  PaintBox1.Refresh;
	  theParentPaintBox.Refresh;
	  END;
	END; {mrAbort:}
      ELSE
	BEGIN
	aTrmLine.Def.bSelected := false;
	DrawLineImmediately( aTrmLine );
	END;
      END; {CASE LineDlg.ShowModal}
    END; {IF aTrmLine <> nil}
  END {mGetWire:}
 ELSE
  END; {CASE theMode OF}

{$IFDEF DEMOVERSION}
  IF theNodes.FNumLines < wMaxDemoLines THEN
    BEGIN
    IF Self = TMDIMainForm(Application.Mainform).ActiveMDIChild THEN
      TMDIMainForm(Application.Mainform).SpeedButtonLTG.Visible 	:= true;
    END;
  {$ENDIF}

END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.FormDestroy(Sender: TObject);

BEGIN
theNodes.Free;
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.SpeedButtonsVisible( bVisible: BOOLEAN );

BEGIN
WITH TMDIMainForm(Application.Mainform) DO
  BEGIN
  SpeedButtonPFNode.Visible	:= bVisible;
  SpeedButtonPFWire.Visible	:= bVisible;

  {$IFDEF DEMOVERSION}
    IF Assigned(theNodes) AND (theNodes.FNumLines < wMaxDemoLines) THEN
      SpeedButtonLTG.Visible 	:= bVisible
     ELSE
      SpeedButtonLTG.Visible 	:= false;
   {$ELSE}
      SpeedButtonLTG.Visible 	:= bVisible;
    {$ENDIF}

  SpeedButtonQuelle.Visible	:= bVisible;
  SpeedButtonZL.Visible 	:= bVisible;
  END;
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.FormActivate(Sender: TObject);

BEGIN
SpeedButtonsVisible( true );
{$IFDEF DEMOVERSION}
  IF Assigned(theNodes) AND (theNodes.FNumLines >= wMaxDemoLines) THEN
    FMode := mNone;
  {$ENDIF}

WITH TMDIMainForm(Application.Mainform) DO
  BEGIN
  SpeedButtonPFNode.Down	:= FMode = mGetNode;
  SpeedButtonPFWire.Down	:= FMode = mGetWire;
  SpeedButtonLTG.Down           := FMode = mDrawTrmLine;
  SpeedButtonQuelle.Down	:= FMode = mPlaceSrc;
  SpeedButtonZL.Down            := FMode = mPlaceZL;
  END;

UpdateMenuItems( Sender );
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.FormDeactivate(Sender: TObject);

BEGIN
SpeedButtonsVisible( false );
END;

(*----------------------------------------------------*)
PROCEDURE TSubForm.Network( aOwner: TComponent );

VAR
  aTNBDlgParameter	: TTNBDlgParameter;
  aStringList		: TStringList;
  strFileName		: TFileName;
  aSaveDialog		: TSaveDialog;

BEGIN
strFileName		:= 'RapidCAN.';
aTNBDlgParameter	:= TTNBDlgParameter.Create( Self );
WITH aTNBDlgParameter DO
  BEGIN
  IF NOT( FileExists( strFileName+'HD' ) ) THEN
    MemoHeader.Lines.SaveToFile( strFileName+'HD' )
   ELSE
    MemoHeader.Lines.LoadFromFile( strFileName+'HD' );
  IF NOT( FileExists( strFileName+'TR' ) ) THEN
    MemoTrailer.Lines.SaveToFile( strFileName+'TR' )
   ELSE
    MemoTrailer.Lines.LoadFromFile( strFileName+'TR' );

  MemoHeader.Lines[0]	:= 'Circuit generated by RapidCAN; (c) U.Creutzburg '+DateTimeToStr(Now);
  aStringList		:= TStringList.Create;
  theNodes.AddCircuitToStringList( aStringList );
  MemoCANLines.Lines.Assign( aStringList );
  aStringList.Free;
  CASE ShowModal OF
    mrOK:
      BEGIN
      MemoHeader.Lines.SaveToFile( strFileName+'HD' );
      MemoCANLines.Lines.SaveToFile( strFileName+'LN' );
      MemoTrailer.Lines.SaveToFile( strFileName+'TR' );
      END;
    4711:
      BEGIN
      MemoHeader.Lines.SaveToFile( strFileName+'HD' );
      MemoCANLines.Lines.SaveToFile( strFileName+'LN' );
      MemoTrailer.Lines.SaveToFile( strFileName+'TR' );
      aSaveDialog	:=  TSaveDialog.Create( aOwner );
      WITH aSaveDialog DO
	BEGIN
	DefaultExt	:= 'CIR';
	Options		:= [ofPathMustExist,ofOverwritePrompt];
	Filter		:= 'PSPICE CIRCUIT|*.cir';
	strFileName	:= theNodes.FFileName;
	Delete( strFileName, Pos('.',strFileName)+1, Length(strFileName)-Pos('.',strFileName) );
	FileName	:= strFileName+'cir';
	InitialDir	:= ExtractFilePath( FileName );
	IF Execute THEN
	  BEGIN
	  aStringList		:= TStringList.Create;
	  aStringList.Assign( aTNBDlgParameter.MemoHeader.Lines );
	  theNodes.AddCircuitToStringList( aStringList );
	  aStringList.AddStrings( aTNBDlgParameter.MemoTrailer.Lines );
	  aStringList.SaveToFile( FileName );
	  aStringList.Free;
	  END; {IF Execute}
	END; {WITH aSaveDialog}
      aSaveDialog.Free;
      END; {4711:}
    END; {CASE ShowModal}
  END; {WITH aTNBDlgParameter}
PaintBox1.Refresh;
aTNBDlgParameter.Free;
END;

(*------------------------------------------------------*)
FUNCTION TSubForm.SaveCANAs( aOwner: TComponent ): TFileName;

VAR
  aSaveDialog	: TSaveDialog;
  strFileName	: TFileName;

BEGIN
Result := '';
aSaveDialog	:=  TSaveDialog.Create( aOwner );
WITH aSaveDialog DO
  BEGIN
  DefaultExt	:= 'RPC';
  Options	:= [ofPathMustExist,ofOverwritePrompt];
  Filter	:= 'RapidCAN|*.rpc';
  strFileName	:= theNodes.FFileName;
  Delete( strFileName, Pos('.',strFileName)+1, Length(strFileName)-Pos('.',strFileName) );
  FileName	:= strFileName+'rpc';
  InitialDir	:= ExtractFilePath( FileName );
  IF Execute THEN
    BEGIN
    IF theNodes.Save( FileName ) THEN
      BEGIN
      Caption := FileName;
      Result  := FileName;
      SendMessage( theParentWindow, CM_ParentCalled, cmPCNewFileName, 0 );
      END;
    END;
  END;
aSaveDialog.Free;
END;

(*------------------------------------------------------*)
FUNCTION  TSubForm.SaveCAN: TFileName;

BEGIN
IF theNodes.Save( theNodes.FFileName ) THEN
  Result := theNodes.FFileName
 ELSE
  Result := '';
END;

(*------------------------------------------------------*)
FUNCTION TSubForm.SaveCircuit( aOwner: TComponent ): TFileName;

VAR
  aSaveDialog	: TSaveDialog;
  strFileName	: TFileName;

BEGIN
Result := '';
aSaveDialog	:=  TSaveDialog.Create( aOwner );
WITH aSaveDialog DO
  BEGIN
  DefaultExt	:= 'CIR';
  Options	:= [ofPathMustExist,ofOverwritePrompt];
  Filter	:= 'PSPICEW CIRCUIT|*.cir';
  strFileName	:= theNodes.FFileName;
  Delete( strFileName, Pos('.',strFileName)+1, Length(strFileName)-Pos('.',strFileName) );
  FileName	:= strFileName+'cir';
  InitialDir	:= ExtractFilePath( FileName );
  IF Execute THEN
    BEGIN
    IF theNodes.SaveCircuit( FileName ) THEN
      Result := FileName;
    END;
  END;
aSaveDialog.Free;
END;

(*------------------------------------------------------*)
FUNCTION TSubForm.UnDo: BOOLEAN;

BEGIN
Result := theNodes.UnDo;
IF Result THEN
  BEGIN
  PaintBox1.Refresh;
  theParentPaintBox.Refresh;
  END;
UpdateMenuItems( Self );
END;

(*------------------------------------------------------*)
PROCEDURE TSubForm.Network1Click(Sender: TObject);

BEGIN
Network( Self );
END;

(*------------------------------------------------------*)
PROCEDURE TSubForm.Undo1Click(Sender: TObject);

BEGIN
UnDo;
END;

(*------------------------------------------------------*)
PROCEDURE TSubForm.Options1Click(Sender: TObject);


VAR
  aPoint : TPoint;

BEGIN
FormOptions := TFormOptions.Create( Self );
WITH FormOptions DO
  BEGIN
  FFontClip      := theParentPaintBox.Font;
  FFontDraw      := PaintBox1.Font;
  EditSizeX.Text := IntToStr(FDimensions.X * theMultiplier.X);
  EditSizeY.Text := IntToStr(FDimensions.Y * theMultiplier.Y);
  EditLineZL.Text:= IntToStr(theDefaults.wLineZL);
  EditZLTerm.Text:= IntToStr(theDefaults.wZL);
  EditNoZL.Text  := IntToStr(theDefaults.wNoZL);
  CheckBoxShowNames.Checked := theDefaults.bShowNodeNames;
  CASE ShowModal OF
    mrOK: BEGIN
	  theDefaults.wLineZL           := StrToIntDef(EditLineZL.Text, theDefaults.wZL);
	  theDefaults.wZL       	:= StrToIntDef(EditZLTerm.Text, theDefaults.wLineZL);
	  theDefaults.wNoZL         	:= StrToIntDef(EditNoZL.Text, theDefaults.wNoZL);
	  theDefaults.bShowNodeNames	:= CheckBoxShowNames.Checked;

	  IF StrToIntDef(EditSizeX.Text, 0) > 0 THEN
	    aPoint.X := StrToIntDef(EditSizeX.Text, 0) DIV theMultiplier.X
	   ELSE
	    aPoint.X := FDimensions.X;
	  IF StrToIntDef(EditSizeY.Text, 0) > 0 THEN
	    aPoint.Y := StrToIntDef(EditSizeY.Text, 0) DIV theMultiplier.Y
	   ELSE
	    aPoint.Y := FDimensions.Y;
	  FDimensions := aPoint;
	  SendMessage( theParentWindow, CM_ParentCalled, cmPCChangeSizeX, FDimensions.X );
	  SendMessage( theParentWindow, CM_ParentCalled, cmPCChangeSizeY, FDimensions.Y );

	  theParentPaintBox.Font := FFontClip;
	  PaintBox1.Font         := FFontDraw;

	  PaintBox1.Refresh;
          theParentPaintBox.Refresh;
          END;
    mrCancel:;
    END;
  END;
FormOptions.Free;
END;

(*------------------------------------------------------*)
PROCEDURE TSubForm.UpdateMenuItems(Sender: TObject);

BEGIN
IF Assigned(theNodes) THEN
  BEGIN
  Undo1.Enabled    := theNodes.FUnDoPossible;
  Network1.Enabled := theNodes.FNetworkExists;
  END;
END;

(*------------------------------------------------------*)
procedure TSubForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);

begin
CanClose := true;
IF (theNodes.FFileState = fstModified) THEN
  BEGIN
  IF MessageDlg( theNodes.FFileName+#10#10+'File was modified, changes will be lost!', mtWarning, [mbOK, mbAbort], 0 ) <> mrOK THEN
    IF Length( SaveCANAs( Self ) ) > 0 THEN
      CanClose := true
     ELSE
      CanClose := false
   ELSE
    CanClose := true;
  END;
END;

(*++++++++++++++++++++++++++++++++++++++++++++++++++++*)
END.
